@@ -17,9 +17,12 @@ if __name__ == "__main__":
         root = client.get_root_node()
         objects = client.get_objects_node()
         struct = client.get_node("ns=2;i=10239")
+        struct_array = client.get_node("ns=2;i=10323")
         before = struct.get_value()
-        client.import_and_register_structures()  # scan server for custom structures and import them
+        before_array = struct_array.get_value()
+        client.load_type_definitions()  # scan server for custom structures and import them
         after = struct.get_value()
+        after_array = struct_array.get_value()
         
 
         embed()
