@@ -1,5 +1,6 @@
 from __future__ import division  # support for python2
 from threading import Thread, Condition
+import concurrent.futures
 import logging
 try:
     from urllib.parse import urlparse
@@ -16,12 +17,12 @@ from opcua.common.subscription import Subscription
 from opcua.common import utils
 from opcua.crypto import security_policies
 from opcua.common.shortcuts import Shortcuts
-from opcua.common.structures_generator import StructGenerator
+from opcua.common.structures import load_type_definitions, load_enums
 use_crypto = True
 try:
     from opcua.crypto import uacrypto
 except ImportError:
-    print("cryptography is not installed, use of crypto disabled")
+    logging.getLogger(__name__).warning("cryptography is not installed, use of crypto disabled")
     use_crypto = False
 
 
@@ -47,7 +48,7 @@ class KeepAlive(Thread):
 
         # some server support no timeout, but we do not trust them
         if self.timeout == 0:
-            self.timeout = 3600000 # 1 hour
+            self.timeout = 3600000  # 1 hour
 
     def run(self):
         self.logger.debug("starting keepalive thread with period of %s milliseconds", self.timeout)
@@ -58,7 +59,11 @@ class KeepAlive(Thread):
             if self._dostop:
                 break
             self.logger.debug("renewing channel")
-            self.client.open_secure_channel(renew=True)
+            try:
+                self.client.open_secure_channel(renew=True)
+            except concurrent.futures.TimeoutError:
+                self.logger.debug("keepalive failed: timeout on open_secure_channel()")
+                break
             val = server_state.get_value()
             self.logger.debug("server state is: %s ", val)
         self.logger.debug("keepalive thread has stopped")
@@ -71,15 +76,14 @@ class KeepAlive(Thread):
 
 
 class Client(object):
-
     """
     High level client to connect to an OPC-UA server.
 
     This class makes it easy to connect and browse address space.
-    It attemps to expose as much functionality as possible
-    but if you want more flexibility it is possible and adviced to
-    use UaClient object, available as self.uaclient
-    which offers the raw OPC-UA services interface.
+    It attempts to expose as much functionality as possible
+    but if you want more flexibility it is possible and advised to
+    use the UaClient object, available as self.uaclient, which offers
+    the raw OPC-UA services interface.
     """
 
     def __init__(self, url, timeout=4):
@@ -95,24 +99,27 @@ class Client(object):
         """
         self.logger = logging.getLogger(__name__)
         self.server_url = urlparse(url)
-        #take initial username and password from the url
+        # take initial username and password from the url
         self._username = self.server_url.username
         self._password = self.server_url.password
         self.name = "Pure Python Client"
         self.description = self.name
         self.application_uri = "urn:freeopcua:client"
-        self.product_uri = "urn:freeopcua.github.no:client"
+        self.product_uri = "urn:freeopcua.github.io:client"
         self.security_policy = ua.SecurityPolicy()
         self.secure_channel_id = None
-        self.secure_channel_timeout = 3600000 # 1 hour
-        self.session_timeout = 3600000 # 1 hour
+        self.secure_channel_timeout = 3600000  # 1 hour
+        self.session_timeout = 3600000  # 1 hour
         self._policy_ids = []
         self.uaclient = UaClient(timeout)
         self.user_certificate = None
         self.user_private_key = None
+        self._server_nonce = None
         self._session_counter = 1
         self.keepalive = None
         self.nodes = Shortcuts(self.uaclient)
+        self.max_messagesize = 0  # No limits
+        self.max_chunkcount = 0  # No limits
 
     def __enter__(self):
         self.connect()
@@ -131,8 +138,7 @@ class Client(object):
                     ep.SecurityMode == security_mode and
                     ep.SecurityPolicyUri == policy_uri):
                 return ep
-        raise ua.UaError("No matching endpoints: {0}, {1}".format(
-                         security_mode, policy_uri))
+        raise ua.UaError("No matching endpoints: {0}, {1}".format(security_mode, policy_uri))
 
     def set_user(self, username):
         """
@@ -152,7 +158,7 @@ class Client(object):
         """
         Set SecureConnection mode. String format:
         Policy,Mode,certificate,private_key[,server_private_key]
-        where Policy is Basic128Rsa15 or Basic256,
+        where Policy is Basic128Rsa15, Basic256 or Basic256Sha256,
             Mode is Sign or SignAndEncrypt
             certificate, private_key and server_private_key are
                 paths to .pem or .der files
@@ -204,11 +210,13 @@ class Client(object):
         Connect, ask server for endpoints, and disconnect
         """
         self.connect_socket()
-        self.send_hello()
-        self.open_secure_channel()
-        endpoints = self.get_endpoints()
-        self.close_secure_channel()
-        self.disconnect_socket()
+        try:
+            self.send_hello()
+            self.open_secure_channel()
+            endpoints = self.get_endpoints()
+            self.close_secure_channel()
+        finally:
+            self.disconnect_socket()
         return endpoints
 
     def connect_and_find_servers(self):
@@ -216,11 +224,13 @@ class Client(object):
         Connect, ask server for a list of known servers, and disconnect
         """
         self.connect_socket()
-        self.send_hello()
-        self.open_secure_channel()  # spec says it should not be necessary to open channel
-        servers = self.find_servers()
-        self.close_secure_channel()
-        self.disconnect_socket()
+        try:
+            self.send_hello()
+            self.open_secure_channel()  # spec says it should not be necessary to open channel
+            servers = self.find_servers()
+            self.close_secure_channel()
+        finally:
+            self.disconnect_socket()
         return servers
 
     def connect_and_find_servers_on_network(self):
@@ -228,11 +238,13 @@ class Client(object):
         Connect, ask server for a list of known servers on network, and disconnect
         """
         self.connect_socket()
-        self.send_hello()
-        self.open_secure_channel()
-        servers = self.find_servers_on_network()
-        self.close_secure_channel()
-        self.disconnect_socket()
+        try:
+            self.send_hello()
+            self.open_secure_channel()
+            servers = self.find_servers_on_network()
+            self.close_secure_channel()
+        finally:
+            self.disconnect_socket()
         return servers
 
     def connect(self):
@@ -241,9 +253,13 @@ class Client(object):
         Connect, create and activate session
         """
         self.connect_socket()
-        self.send_hello()
-        self.open_secure_channel()
-        self.create_session()
+        try:
+            self.send_hello()
+            self.open_secure_channel()
+            self.create_session()
+        except Exception:
+            self.disconnect_socket()  # clean up open socket
+            raise
         self.activate_session(username=self._username, password=self._password, certificate=self.user_certificate)
 
     def disconnect(self):
@@ -251,9 +267,11 @@ class Client(object):
         High level method
         Close session, secure channel and socket
         """
-        self.close_session()
-        self.close_secure_channel()
-        self.disconnect_socket()
+        try:
+            self.close_session()
+            self.close_secure_channel()
+        finally:
+            self.disconnect_socket()
 
     def connect_socket(self):
         """
@@ -268,8 +286,11 @@ class Client(object):
         """
         Send OPC-UA hello to server
         """
-        ack = self.uaclient.send_hello(self.server_url.geturl())
-        # FIXME check ack
+        ack = self.uaclient.send_hello(self.server_url.geturl(), self.max_messagesize, self.max_chunkcount)
+
+        # TODO: Handle ua.UaError
+        if isinstance(ack, ua.UaStatusCodeError):
+            raise ack
 
     def open_secure_channel(self, renew=False):
         """
@@ -282,8 +303,9 @@ class Client(object):
             params.RequestType = ua.SecurityTokenRequestType.Renew
         params.SecurityMode = self.security_policy.Mode
         params.RequestedLifetime = self.secure_channel_timeout
-        nonce = utils.create_nonce(self.security_policy.symmetric_key_size)   # length should be equal to the length of key of symmetric encryption
-        params.ClientNonce = nonce	# this nonce is used to create a symmetric key
+        # length should be equal to the length of key of symmetric encryption
+        nonce = utils.create_nonce(self.security_policy.symmetric_key_size)
+        params.ClientNonce = nonce  # this nonce is used to create a symmetric key
         result = self.uaclient.open_secure_channel(params)
         self.security_policy.make_symmetric_key(nonce, result.ServerNonce)
         self.secure_channel_timeout = result.SecurityToken.RevisedLifetime
@@ -296,26 +318,6 @@ class Client(object):
         params.EndpointUrl = self.server_url.geturl()
         return self.uaclient.get_endpoints(params)
 
-    def register_server(self, server, discovery_configuration=None):
-        """
-        register a server to discovery server
-        if discovery_configuration is provided, the newer register_server2 service call is used
-        """
-        serv = ua.RegisteredServer()
-        serv.ServerUri = server.application_uri
-        serv.ProductUri = server.product_uri
-        serv.DiscoveryUrls = [server.endpoint.geturl()]
-        serv.ServerType = server.application_type
-        serv.ServerNames = [ua.LocalizedText(server.name)]
-        serv.IsOnline = True
-        if discovery_configuration:
-            params = ua.RegisterServer2Parameters()
-            params.Server = serv
-            params.DiscoveryConfiguration = discovery_configuration
-            return self.uaclient.register_server2(params)
-        else:
-            return self.uaclient.register_server(serv)
-
     def find_servers(self, uris=None):
         """
         send a FindServer request to the server. The answer should be a list of
@@ -346,7 +348,8 @@ class Client(object):
         desc.ApplicationType = ua.ApplicationType.Client
 
         params = ua.CreateSessionParameters()
-        nonce = utils.create_nonce(32)  # at least 32 random bytes for server to prove possession of private key (specs part 4, 5.6.2.2)
+        # at least 32 random bytes for server to prove possession of private key (specs part 4, 5.6.2.2)
+        nonce = utils.create_nonce(32)
         params.ClientNonce = nonce
         params.ClientCertificate = self.security_policy.client_certificate
         params.ClientDescription = desc
@@ -369,7 +372,8 @@ class Client(object):
         ep = Client.find_endpoint(response.ServerEndpoints, self.security_policy.Mode, self.security_policy.URI)
         self._policy_ids = ep.UserIdentityTokens
         self.session_timeout = response.RevisedSessionTimeout
-        self.keepalive = KeepAlive(self, min(self.session_timeout, self.secure_channel_timeout) * 0.7)  # 0.7 is from spec
+        self.keepalive = KeepAlive(
+            self, min(self.session_timeout, self.secure_channel_timeout) * 0.7)  # 0.7 is from spec
         self.keepalive.start()
         return response
 
@@ -407,7 +411,7 @@ class Client(object):
             challenge += self.security_policy.server_certificate
         if self._server_nonce is not None:
             challenge += self._server_nonce
-        params.ClientSignature.Algorithm = b"http://www.w3.org/2000/09/xmldsig#rsa-sha1"
+        params.ClientSignature.Algorithm = "http://www.w3.org/2000/09/xmldsig#rsa-sha1"
         params.ClientSignature.Signature = self.security_policy.asymmetric_cryptography.signature(challenge)
         params.LocaleIds.append("en")
         if not username and not certificate:
@@ -420,17 +424,17 @@ class Client(object):
 
     def _add_anonymous_auth(self, params):
         params.UserIdentityToken = ua.AnonymousIdentityToken()
-        params.UserIdentityToken.PolicyId = self.server_policy_id(ua.UserTokenType.Anonymous, b"anonymous")
+        params.UserIdentityToken.PolicyId = self.server_policy_id(ua.UserTokenType.Anonymous, "anonymous")
 
     def _add_certificate_auth(self, params, certificate, challenge):
         params.UserIdentityToken = ua.X509IdentityToken()
-        params.UserIdentityToken.PolicyId = self.server_policy_id(ua.UserTokenType.Certificate, b"certificate_basic256")
+        params.UserIdentityToken.PolicyId = self.server_policy_id(ua.UserTokenType.Certificate, "certificate_basic256")
         params.UserIdentityToken.CertificateData = uacrypto.der_from_x509(certificate)
         # specs part 4, 5.6.3.1: the data to sign is created by appending
         # the last serverNonce to the serverCertificate
         sig = uacrypto.sign_sha1(self.user_private_key, challenge)
         params.UserTokenSignature = ua.SignatureData()
-        params.UserTokenSignature.Algorithm = b"http://www.w3.org/2000/09/xmldsig#rsa-sha1"
+        params.UserTokenSignature.Algorithm = "http://www.w3.org/2000/09/xmldsig#rsa-sha1"
         params.UserTokenSignature.Signature = sig
 
     def _add_user_auth(self, params, username, password):
@@ -443,13 +447,13 @@ class Client(object):
             # and EncryptionAlgorithm is null
             if self._password:
                 self.logger.warning("Sending plain-text password")
-                params.UserIdentityToken.Password = password
-            params.UserIdentityToken.EncryptionAlgorithm = ''
+                params.UserIdentityToken.Password = password.encode('utf8')
+            params.UserIdentityToken.EncryptionAlgorithm = None
         elif self._password:
             data, uri = self._encrypt_password(password, policy_uri)
             params.UserIdentityToken.Password = data
             params.UserIdentityToken.EncryptionAlgorithm = uri
-        params.UserIdentityToken.PolicyId = self.server_policy_id(ua.UserTokenType.UserName, b"username_basic256")
+        params.UserIdentityToken.PolicyId = self.server_policy_id(ua.UserTokenType.UserName, "username_basic256")
 
     def _encrypt_password(self, password, policy_uri):
         pubkey = uacrypto.x509_from_der(self.security_policy.server_certificate).public_key()
@@ -466,8 +470,9 @@ class Client(object):
         """
         Close session
         """
-        if self.keepalive:
+        if self.keepalive and self.keepalive.is_alive():
             self.keepalive.stop()
+            self.keepalive.join()
         return self.uaclient.close_session(True)
 
     def get_root_node(self):
@@ -491,7 +496,7 @@ class Client(object):
         returns a Subscription object which allow
         to subscribe to events or data on server
         handler argument is a class with data_change and/or event methods.
-        period argument is either a publishing interval in seconds or a
+        period argument is either a publishing interval in milliseconds or a
         CreateSubscriptionParameters instance. The second option should be used,
         if the opcua-server has problems with the default options.
         These methods will be called when notfication from server are received.
@@ -502,7 +507,7 @@ class Client(object):
         """
 
         if isinstance(period, ua.CreateSubscriptionParameters):
-             return Subscription(self.uaclient, period, handler)
+            return Subscription(self.uaclient, period, handler)
         params = ua.CreateSubscriptionParameters()
         params.RequestedPublishingInterval = period
         params.RequestedLifetimeCount = 10000
@@ -523,12 +528,12 @@ class Client(object):
     def delete_nodes(self, nodes, recursive=False):
         return delete_nodes(self.uaclient, nodes, recursive)
 
-    def import_xml(self, path):
+    def import_xml(self, path=None, xmlstring=None):
         """
         Import nodes defined in xml
         """
         importer = XmlImporter(self)
-        return importer.import_xml(path)
+        return importer.import_xml(path, xmlstring)
 
     def export_xml(self, nodes, path):
         """
@@ -551,45 +556,17 @@ class Client(object):
         ns_node.set_value(uries)
         return len(uries) - 1
 
-    def import_and_register_structures(self, nodes=None):
+    def load_type_definitions(self, nodes=None):
         """
-        Download xml from given variable node defining custom structures.
-        If no no node is given, attemps to import variables from all nodes under
-        "0:OPC Binary"
-        the code is generated and imported on the fly. If you know the structures
-        are not going to be modified it might be interresting to copy the generated files
-        and include them in you code
+        Load custom types (custom structures/extension objects) definition from server
+        Generate Python classes for custom structures/extension objects defined in server
+        These classes will available in ua module
         """
-        if nodes is None:
-            nodes = []
-            for desc in self.nodes.opc_binary.get_children_descriptions():
-                if desc.BrowseName != ua.QualifiedName("Opc.Ua"):
-                    nodes.append(self.get_node(desc.NodeId))
-        self.logger.info("Importing structures from nodes: %s", nodes)
-        
-        structs_dict = {}
-        for node in nodes:
-            xml = node.get_value()
-            xml = xml.decode("utf-8")
-            name = "structures_" + node.get_browse_name().Name
-            gen = StructGenerator()
-            gen.make_model_from_string(xml)
-            gen.save_and_import(name + ".py", append_to=structs_dict)
+        return load_type_definitions(self, nodes)
 
-        # register classes
-        for desc in self.nodes.base_structure_type.get_children_descriptions():
-            # FIXME: maybe we should look recursively at children
-            # FIXME: we should get enoding and description but this is too 
-            # expensive. we take a shorcut and assume that browsename of struct 
-            # is the same as the name of the data type structure
-            if desc.BrowseName.Name in structs_dict:
-                struct_node = self.get_node(desc.NodeId)
-                refs = struct_node.get_references(ua.ObjectIds.HasEncoding, ua.BrowseDirection.Forward)
-                for ref in refs:
-                    if "Binary" in ref.BrowseName.Name:
-                        ua.register_extension_object(desc.BrowseName.Name, ref.NodeId, structs_dict[desc.BrowseName.Name])
-
-
-
-
-            
+    def load_enums(self):
+        """
+        generate Python enums for custom enums on server.
+        This enums will be available in ua module
+        """
+        return load_enums(self)
