@@ -4,13 +4,15 @@ Low level binary client
 
 import logging
 import socket
+import errno
 from threading import Thread, Lock
 from concurrent.futures import Future
 from functools import partial
 
 from opcua import ua
-from opcua.common import utils
+from opcua.ua.ua_binary import struct_from_binary, uatcp_to_binary, struct_to_binary, nodeid_from_binary
 from opcua.ua.uaerrors import UaError, BadTimeout, BadNoSubscription, BadSessionClosed
+from opcua.common.connection import SecureConnection
 
 
 class UASocketClient(object):
@@ -29,7 +31,7 @@ class UASocketClient(object):
         self._request_id = 0
         self._request_handle = 0
         self._callbackmap = {}
-        self._connection = ua.SecureConnection(security_policy)
+        self._connection = SecureConnection(security_policy)
 
     def start(self):
         """
@@ -50,8 +52,8 @@ class UASocketClient(object):
             request.RequestHeader = self._create_request_header(timeout)
             self.logger.debug("Sending: %s", request)
             try:
-                binreq = request.to_binary()
-            except:
+                binreq = struct_to_binary(request)
+            except Exception:
                 # reset reqeust handle if any error
                 # see self._create_request_header
                 self._request_handle -= 1
@@ -79,10 +81,10 @@ class UASocketClient(object):
 
     def check_answer(self, data, context):
         data = data.copy()
-        typeid = ua.NodeId.from_binary(data)
+        typeid = nodeid_from_binary(data)
         if typeid == ua.FourByteNodeId(ua.ObjectIds.ServiceFault_Encoding_DefaultBinary):
             self.logger.warning("ServiceFault from server received %s", context)
-            hdr = ua.ResponseHeader.from_binary(data)
+            hdr = struct_from_binary(ua.ResponseHeader, data)
             hdr.ServiceResult.check()
             return False
         return True
@@ -108,7 +110,8 @@ class UASocketClient(object):
         elif isinstance(msg, ua.Acknowledge):
             self._call_callback(0, msg)
         elif isinstance(msg, ua.ErrorMessage):
-            self.logger.warning("Received an error: %s", msg)
+            self.logger.fatal("Received an error: %s", msg)
+            self._call_callback(0, ua.UaStatusCodeError(msg.Error.value))
         else:
             raise ua.UaError("Unsupported message type: %s", msg)
 
@@ -116,7 +119,10 @@ class UASocketClient(object):
         with self._lock:
             future = self._callbackmap.pop(request_id, None)
             if future is None:
-                raise ua.UaError("No future object found for request: {0}, callbacks in list are {1}".format(request_id, self._callbackmap.keys()))
+                raise ua.UaError(
+                    "No future object found for request: {0}, callbacks in list are {1}"
+                    .format(request_id, self._callbackmap.keys())
+                )
         future.set_result(body)
 
     def _create_request_header(self, timeout=1000):
@@ -132,24 +138,41 @@ class UASocketClient(object):
         connect to server socket and start receiving thread
         """
         self.logger.info("opening connection")
-        sock = socket.create_connection((host, port))
-        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)  # nodelay ncessary to avoid packing in one frame, some servers do not like it
-        self._socket = utils.SocketWrapper(sock)
+        # Create socket with timeout for initial connection
+        sock = socket.create_connection((host, port), timeout=self.timeout)
+        # set to blocking mode again
+        sock.settimeout(None)
+        # nodelay necessary to avoid packing in one frame, some servers do not like it
+        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
+        self._socket = ua.utils.SocketWrapper(sock)
         self.start()
 
     def disconnect_socket(self):
-        self.logger.info("stop request")
+        self.logger.info("Request to close socket received")
         self._do_stop = True
-        self._socket.socket.shutdown(socket.SHUT_RDWR)
+        try:
+            self._socket.socket.shutdown(socket.SHUT_RDWR)
+        except (socket.error, OSError) as exc:
+            if exc.errno in (errno.ENOTCONN, errno.EBADF):
+                pass # Socket is not connected, so can't send FIN packet.
+            else:
+                raise
         self._socket.socket.close()
+        self.logger.info("Socket closed, waiting for receiver thread to terminate...")
+        if self._thread and self._thread.is_alive():
+            self._thread.join()
+        self._callbackmap.clear()
+        self.logger.info("Done closing socket: Receiving thread terminated, socket disconnected")
 
-    def send_hello(self, url):
+    def send_hello(self, url, max_messagesize=0, max_chunkcount=0):
         hello = ua.Hello()
         hello.EndpointUrl = url
+        hello.MaxMessageSize = max_messagesize
+        hello.MaxChunkCount = max_chunkcount
         future = Future()
         with self._lock:
             self._callbackmap[0] = future
-        binmsg = self._connection.tcp_to_binary(ua.MessageType.Hello, hello)
+        binmsg = uatcp_to_binary(ua.MessageType.Hello, hello)
         self._socket.write(binmsg)
         ack = future.result(self.timeout)
         return ack
@@ -159,29 +182,36 @@ class UASocketClient(object):
         request = ua.OpenSecureChannelRequest()
         request.Parameters = params
         future = self._send_request(request, message_type=ua.MessageType.SecureOpen)
-        
+
         # FIXME: we have a race condition here
         # we can get a packet with the new token id before we reach to store it..
-        response = ua.OpenSecureChannelResponse.from_binary(future.result(self.timeout))
+        response = struct_from_binary(ua.OpenSecureChannelResponse, future.result(self.timeout))
         response.ResponseHeader.ServiceResult.check()
         self._connection.set_channel(response.Parameters)
         return response.Parameters
 
     def close_secure_channel(self):
         """
-        close secure channel. It seems to trigger a shutdown of socket
-        in most servers, so be prepare to reconnect.
-        OPC UA specs Part 6, 7.1.4 say that Server does not send a CloseSecureChannel response and should just close socket
+        close secure channel. It seems to trigger a shutdown of socket in most servers, so be prepare to reconnect.
+        OPC UA specs Part 6, 7.1.4 say that Server does not send a CloseSecureChannel response and should just close
+        socket
         """
         self.logger.info("close_secure_channel")
         request = ua.CloseSecureChannelRequest()
-        future = self._send_request(request, message_type=ua.MessageType.SecureClose)
-        with self._lock:
-            # don't expect any more answers
-            future.cancel()
-            self._callbackmap.clear()
+        try:
+            future = self._send_request(request, message_type=ua.MessageType.SecureClose)
+            with self._lock:
+                # some servers send a response here, most do not ... so we ignore
+                future.cancel()
+        except (socket.error, OSError) as exc:
+            if exc.errno in (errno.ENOTCONN, errno.EBADF):
+                # Socket is closed, so can't send CloseSecureChannelRequest.
+                self.logger.warning("close_secure_channel() failed: socket already closed")
+            else:
+                raise
 
-        # some servers send a response here, most do not ... so we ignore
+    def is_secure_channel_open(self):
+        return self._connection.is_open()
 
 
 class UaClient(object):
@@ -202,23 +232,23 @@ class UaClient(object):
         self._publishcallbacks = {}
         self._timeout = timeout
         self._uasocket = None
-        self._security_policy = ua.SecurityPolicy()
+        self.security_policy = ua.SecurityPolicy()
 
     def set_security(self, policy):
-        self._security_policy = policy
+        self.security_policy = policy
 
     def connect_socket(self, host, port):
         """
         connect to server socket and start receiving thread
         """
-        self._uasocket = UASocketClient(self._timeout, security_policy=self._security_policy)
+        self._uasocket = UASocketClient(self._timeout, security_policy=self.security_policy)
         return self._uasocket.connect_socket(host, port)
 
     def disconnect_socket(self):
         return self._uasocket.disconnect_socket()
 
-    def send_hello(self, url):
-        return self._uasocket.send_hello(url)
+    def send_hello(self, url, max_messagesize=0, max_chunkcount=0):
+        return self._uasocket.send_hello(url, max_messagesize, max_chunkcount)
 
     def open_secure_channel(self, params):
         return self._uasocket.open_secure_channel(params)
@@ -235,7 +265,7 @@ class UaClient(object):
         request = ua.CreateSessionRequest()
         request.Parameters = parameters
         data = self._uasocket.send_request(request)
-        response = ua.CreateSessionResponse.from_binary(data)
+        response = struct_from_binary(ua.CreateSessionResponse, data)
         self.logger.debug(response)
         response.ResponseHeader.ServiceResult.check()
         self._uasocket.authentication_token = response.Parameters.AuthenticationToken
@@ -246,17 +276,20 @@ class UaClient(object):
         request = ua.ActivateSessionRequest()
         request.Parameters = parameters
         data = self._uasocket.send_request(request)
-        response = ua.ActivateSessionResponse.from_binary(data)
+        response = struct_from_binary(ua.ActivateSessionResponse, data)
         self.logger.debug(response)
         response.ResponseHeader.ServiceResult.check()
         return response.Parameters
 
     def close_session(self, deletesubscriptions):
         self.logger.info("close_session")
+        # Bail out if we don't have an open server-channel to unsubscribe ourself.
+        if not self._uasocket.is_secure_channel_open():
+            return
         request = ua.CloseSessionRequest()
         request.DeleteSubscriptions = deletesubscriptions
         data = self._uasocket.send_request(request)
-        response = ua.CloseSessionResponse.from_binary(data)
+        response = struct_from_binary(ua.CloseSessionResponse, data)
         try:
             response.ResponseHeader.ServiceResult.check()
         except BadSessionClosed:
@@ -271,7 +304,7 @@ class UaClient(object):
         request = ua.BrowseRequest()
         request.Parameters = parameters
         data = self._uasocket.send_request(request)
-        response = ua.BrowseResponse.from_binary(data)
+        response = struct_from_binary(ua.BrowseResponse, data)
         self.logger.debug(response)
         response.ResponseHeader.ServiceResult.check()
         return response.Results
@@ -281,7 +314,7 @@ class UaClient(object):
         request = ua.BrowseNextRequest()
         request.Parameters = parameters
         data = self._uasocket.send_request(request)
-        response = ua.BrowseNextResponse.from_binary(data)
+        response = struct_from_binary(ua.BrowseNextResponse, data)
         self.logger.debug(response)
         response.ResponseHeader.ServiceResult.check()
         return response.Parameters.Results
@@ -291,7 +324,7 @@ class UaClient(object):
         request = ua.ReadRequest()
         request.Parameters = parameters
         data = self._uasocket.send_request(request)
-        response = ua.ReadResponse.from_binary(data)
+        response = struct_from_binary(ua.ReadResponse, data)
         self.logger.debug(response)
         response.ResponseHeader.ServiceResult.check()
         # cast to Enum attributes that need to
@@ -311,7 +344,7 @@ class UaClient(object):
         request = ua.WriteRequest()
         request.Parameters = params
         data = self._uasocket.send_request(request)
-        response = ua.WriteResponse.from_binary(data)
+        response = struct_from_binary(ua.WriteResponse, data)
         self.logger.debug(response)
         response.ResponseHeader.ServiceResult.check()
         return response.Results
@@ -321,7 +354,7 @@ class UaClient(object):
         request = ua.GetEndpointsRequest()
         request.Parameters = params
         data = self._uasocket.send_request(request)
-        response = ua.GetEndpointsResponse.from_binary(data)
+        response = struct_from_binary(ua.GetEndpointsResponse, data)
         self.logger.debug(response)
         response.ResponseHeader.ServiceResult.check()
         return response.Endpoints
@@ -331,7 +364,7 @@ class UaClient(object):
         request = ua.FindServersRequest()
         request.Parameters = params
         data = self._uasocket.send_request(request)
-        response = ua.FindServersResponse.from_binary(data)
+        response = struct_from_binary(ua.FindServersResponse, data)
         self.logger.debug(response)
         response.ResponseHeader.ServiceResult.check()
         return response.Servers
@@ -341,7 +374,7 @@ class UaClient(object):
         request = ua.FindServersOnNetworkRequest()
         request.Parameters = params
         data = self._uasocket.send_request(request)
-        response = ua.FindServersOnNetworkResponse.from_binary(data)
+        response = struct_from_binary(ua.FindServersOnNetworkResponse, data)
         self.logger.debug(response)
         response.ResponseHeader.ServiceResult.check()
         return response.Parameters
@@ -351,7 +384,7 @@ class UaClient(object):
         request = ua.RegisterServerRequest()
         request.Server = registered_server
         data = self._uasocket.send_request(request)
-        response = ua.RegisterServerResponse.from_binary(data)
+        response = struct_from_binary(ua.RegisterServerResponse, data)
         self.logger.debug(response)
         response.ResponseHeader.ServiceResult.check()
         # nothing to return for this service
@@ -361,7 +394,7 @@ class UaClient(object):
         request = ua.RegisterServer2Request()
         request.Parameters = params
         data = self._uasocket.send_request(request)
-        response = ua.RegisterServer2Response.from_binary(data)
+        response = struct_from_binary(ua.RegisterServer2Response, data)
         self.logger.debug(response)
         response.ResponseHeader.ServiceResult.check()
         return response.ConfigurationResults
@@ -371,7 +404,7 @@ class UaClient(object):
         request = ua.TranslateBrowsePathsToNodeIdsRequest()
         request.Parameters.BrowsePaths = browsepaths
         data = self._uasocket.send_request(request)
-        response = ua.TranslateBrowsePathsToNodeIdsResponse.from_binary(data)
+        response = struct_from_binary(ua.TranslateBrowsePathsToNodeIdsResponse, data)
         self.logger.debug(response)
         response.ResponseHeader.ServiceResult.check()
         return response.Results
@@ -388,7 +421,7 @@ class UaClient(object):
     def _create_subscription_callback(self, pub_callback, resp_fut, data_fut):
         self.logger.info("_create_subscription_callback")
         data = data_fut.result()
-        response = ua.CreateSubscriptionResponse.from_binary(data)
+        response = struct_from_binary(ua.CreateSubscriptionResponse, data)
         self.logger.debug(response)
         response.ResponseHeader.ServiceResult.check()
         self._publishcallbacks[response.Parameters.SubscriptionId] = pub_callback
@@ -406,7 +439,7 @@ class UaClient(object):
     def _delete_subscriptions_callback(self, subscriptionids, resp_fut, data_fut):
         self.logger.info("_delete_subscriptions_callback")
         data = data_fut.result()
-        response = ua.DeleteSubscriptionsResponse.from_binary(data)
+        response = struct_from_binary(ua.DeleteSubscriptionsResponse, data)
         self.logger.debug(response)
         response.ResponseHeader.ServiceResult.check()
         for sid in subscriptionids:
@@ -419,8 +452,7 @@ class UaClient(object):
             acks = []
         request = ua.PublishRequest()
         request.Parameters.SubscriptionAcknowledgements = acks
-        # timeout could be set to 0 (= no timeout) but some servers do not support it
-        self._uasocket.send_request(request, self._call_publish_callback, timeout=int(9e8)) # 250 days
+        self._uasocket.send_request(request, self._call_publish_callback, timeout=0)
 
     def _call_publish_callback(self, future):
         self.logger.info("call_publish_callback")
@@ -429,10 +461,10 @@ class UaClient(object):
         # check if answer looks ok
         try:
             self._uasocket.check_answer(data, "while waiting for publish response")
-        except BadTimeout: # Spec Part 4, 7.28
+        except BadTimeout:  # Spec Part 4, 7.28
             self.publish()
             return
-        except BadNoSubscription: # Spec Part 5, 13.8.1
+        except BadNoSubscription:  # Spec Part 5, 13.8.1
             # BadNoSubscription is expected after deleting the last subscription.
             #
             # We should therefore also check for len(self._publishcallbacks) == 0, but
@@ -451,14 +483,14 @@ class UaClient(object):
 
         # parse publish response
         try:
-            response = ua.PublishResponse.from_binary(data)
+            response = struct_from_binary(ua.PublishResponse, data)
             self.logger.debug(response)
         except Exception:
             # INFO: catching the exception here might be obsolete because we already
             #       catch BadTimeout above. However, it's not really clear what this code
             #       does so it stays in, doesn't seem to hurt.
             self.logger.exception("Error parsing notificatipn from server")
-            self.publish([]) #send publish request ot server so he does stop sending notifications
+            self.publish([])  # send publish request ot server so he does stop sending notifications
             return
 
         # look for callback
@@ -479,7 +511,7 @@ class UaClient(object):
         request = ua.CreateMonitoredItemsRequest()
         request.Parameters = params
         data = self._uasocket.send_request(request)
-        response = ua.CreateMonitoredItemsResponse.from_binary(data)
+        response = struct_from_binary(ua.CreateMonitoredItemsResponse, data)
         self.logger.debug(response)
         response.ResponseHeader.ServiceResult.check()
         return response.Results
@@ -489,7 +521,7 @@ class UaClient(object):
         request = ua.DeleteMonitoredItemsRequest()
         request.Parameters = params
         data = self._uasocket.send_request(request)
-        response = ua.DeleteMonitoredItemsResponse.from_binary(data)
+        response = struct_from_binary(ua.DeleteMonitoredItemsResponse, data)
         self.logger.debug(response)
         response.ResponseHeader.ServiceResult.check()
         return response.Results
@@ -499,7 +531,7 @@ class UaClient(object):
         request = ua.AddNodesRequest()
         request.Parameters.NodesToAdd = nodestoadd
         data = self._uasocket.send_request(request)
-        response = ua.AddNodesResponse.from_binary(data)
+        response = struct_from_binary(ua.AddNodesResponse, data)
         self.logger.debug(response)
         response.ResponseHeader.ServiceResult.check()
         return response.Results
@@ -509,17 +541,27 @@ class UaClient(object):
         request = ua.AddReferencesRequest()
         request.Parameters.ReferencesToAdd = refs
         data = self._uasocket.send_request(request)
-        response = ua.AddReferencesResponse.from_binary(data)
+        response = struct_from_binary(ua.AddReferencesResponse, data)
         self.logger.debug(response)
         response.ResponseHeader.ServiceResult.check()
         return response.Results
 
+    def delete_references(self, refs):
+        self.logger.info("delete")
+        request = ua.DeleteReferencesRequest()
+        request.Parameters.ReferencesToDelete = refs
+        data = self._uasocket.send_request(request)
+        response = struct_from_binary(ua.DeleteReferencesResponse, data)
+        self.logger.debug(response)
+        response.ResponseHeader.ServiceResult.check()
+        return response.Parameters.Results
+
     def delete_nodes(self, params):
         self.logger.info("delete_nodes")
         request = ua.DeleteNodesRequest()
         request.Parameters = params
         data = self._uasocket.send_request(request)
-        response = ua.DeleteNodesResponse.from_binary(data)
+        response = struct_from_binary(ua.DeleteNodesResponse, data)
         self.logger.debug(response)
         response.ResponseHeader.ServiceResult.check()
         return response.Results
@@ -528,7 +570,7 @@ class UaClient(object):
         request = ua.CallRequest()
         request.Parameters.MethodsToCall = methodstocall
         data = self._uasocket.send_request(request)
-        response = ua.CallResponse.from_binary(data)
+        response = struct_from_binary(ua.CallResponse, data)
         self.logger.debug(response)
         response.ResponseHeader.ServiceResult.check()
         return response.Results
@@ -538,7 +580,7 @@ class UaClient(object):
         request = ua.HistoryReadRequest()
         request.Parameters = params
         data = self._uasocket.send_request(request)
-        response = ua.HistoryReadResponse.from_binary(data)
+        response = struct_from_binary(ua.HistoryReadResponse, data)
         self.logger.debug(response)
         response.ResponseHeader.ServiceResult.check()
         return response.Results
@@ -548,7 +590,7 @@ class UaClient(object):
         request = ua.ModifyMonitoredItemsRequest()
         request.Parameters = params
         data = self._uasocket.send_request(request)
-        response = ua.ModifyMonitoredItemsResponse.from_binary(data)
+        response = struct_from_binary(ua.ModifyMonitoredItemsResponse, data)
         self.logger.debug(response)
         response.ResponseHeader.ServiceResult.check()
         return response.Results
