@@ -2,11 +2,32 @@
 High level node object, to access node attribute
 and browse address space
 """
+from datetime import datetime
 
 from opcua import ua
 from opcua.common import events
 import opcua.common
 
+
+def _check_results(results, reqlen=1):
+    assert len(results) == reqlen, results
+    for r in results:
+        r.check()
+
+
+def _to_nodeid(nodeid):
+    if isinstance(nodeid, int):
+        return ua.TwoByteNodeId(nodeid)
+    elif isinstance(nodeid, Node):
+        return nodeid.nodeid
+    elif isinstance(nodeid, ua.NodeId):
+        return nodeid
+    elif type(nodeid) in (str, bytes):
+        return ua.NodeId.from_string(nodeid)
+    else:
+        raise ua.UaError("Could not resolve '{0}' to a type id".format(nodeid))
+
+
 class Node(object):
 
     """
@@ -125,6 +146,8 @@ class Node(object):
         """
         Get value of a node as a python type. Only variables ( and properties) have values.
         An exception will be generated for other node types.
+        WARNING: on server side, this function returns a ref to object in ua database. Do not modify it if it is a mutable
+        object unless you know what you are doing
         """
         result = self.get_data_value()
         return result.Value.Value
@@ -175,14 +198,19 @@ class Node(object):
         optionnaly using the variantype argument.
         * a ua.Variant, varianttype is then ignored
         * a ua.DataValue, you then have full control over data send to server
+        WARNING: On server side, ref to object is directly saved in our UA db, if this is a mutable object
+        and you modfy it afterward, then the object in db will be modified without any
+        data change event generated
         """
         datavalue = None
         if isinstance(value, ua.DataValue):
             datavalue = value
         elif isinstance(value, ua.Variant):
             datavalue = ua.DataValue(value)
+            datavalue.SourceTimestamp = datetime.utcnow()
         else:
             datavalue = ua.DataValue(ua.Variant(value, varianttype))
+            datavalue.SourceTimestamp = datetime.utcnow()
         self.set_attribute(ua.AttributeIds.Value, datavalue)
 
     set_data_value = set_value
@@ -325,7 +353,7 @@ class Node(object):
         """
         desc = ua.BrowseDescription()
         desc.BrowseDirection = direction
-        desc.ReferenceTypeId = ua.TwoByteNodeId(refs)
+        desc.ReferenceTypeId = _to_nodeid(refs)
         desc.IncludeSubtypes = includesubtypes
         desc.NodeClassMask = nodeclassmask
         desc.ResultMask = ua.BrowseResultMask.All
@@ -372,21 +400,7 @@ class Node(object):
             return None
         return references[0].NodeId
 
-    def get_path_as_string(self, max_length=20):
-        """
-        Attempt to find path of node from root node and return it as a list of strings.
-        There might several possible paths to a node, this function will return one
-        Some nodes may be missing references, so this method may
-        return an empty list
-        Since address space may have circular references, a max length is specified
-
-        """
-        path = self._get_path(max_length)
-        path = [ref.BrowseName.to_string() for ref in path]
-        path.append(self.get_browse_name().to_string())
-        return path
-
-    def get_path(self, max_length=20):
+    def get_path(self, max_length=20, as_string=False):
         """
         Attempt to find path of node from root node and return it as a list of Nodes.
         There might several possible paths to a node, this function will return one
@@ -398,6 +412,8 @@ class Node(object):
         path = self._get_path(max_length)
         path = [Node(self.server, ref.NodeId) for ref in path]
         path.append(self)
+        if as_string:
+            path = [el.get_browse_name().to_string() for el in path]
         return path
 
     def _get_path(self, max_length=20):
@@ -558,17 +574,80 @@ class Node(object):
         result = self.server.history_read(params)[0]
         return result
 
-    def delete(self, delete_references=True):
+    def delete(self, delete_references=True, recursive=False):
         """
         Delete node from address space
         """
-        ditem = ua.DeleteNodesItem()
-        ditem.NodeId = self.nodeid
-        ditem.DeleteTargetReferences = delete_references
-        params = ua.DeleteNodesParameters()
-        params.NodesToDelete = [ditem]
-        result = self.server.delete_nodes(params)
-        result[0].check()
+        nodes, results = opcua.common.manage_nodes.delete_nodes(self.server, [self], recursive, delete_references)
+        for r in results:
+            r.check()
+        return nodes
+
+    def _fill_delete_reference_item(self, rdesc, bidirectional = False):
+        ditem = ua.DeleteReferencesItem()
+        ditem.SourceNodeId = self.nodeid
+        ditem.TargetNodeId = rdesc.NodeId
+        ditem.ReferenceTypeId = rdesc.ReferenceTypeId
+        ditem.IsForward = rdesc.IsForward
+        ditem.DeleteBidirectional = bidirectional
+        return ditem
+
+    def delete_reference(self, target, reftype, forward=True, bidirectional=True):
+        """
+        Delete given node's references from address space
+        """
+        known_refs = self.get_references(reftype, includesubtypes=False)
+        targetid = _to_nodeid(target)
+
+        for r in known_refs:
+            if r.NodeId == targetid and r.IsForward == forward:
+                rdesc = r
+                break
+        else:
+            raise ua.UaStatusCodeError(ua.StatusCodes.BadNotFound)
+
+        ditem = self._fill_delete_reference_item(rdesc, bidirectional)
+        self.server.delete_references([ditem])[0].check()
+
+    def add_reference(self, target, reftype, forward=True, bidirectional=True):
+        """
+        Add reference to node
+        """
+
+        aitem = ua.AddReferencesItem()
+        aitem.SourceNodeId = self.nodeid
+        aitem.TargetNodeId = _to_nodeid(target)
+        aitem.ReferenceTypeId = _to_nodeid(reftype)
+        aitem.IsForward = forward
+
+        params = [aitem]
+
+        if bidirectional:
+            aitem2 = ua.AddReferencesItem()
+            aitem2.SourceNodeId = aitem.TargetNodeId
+            aitem2.TargetNodeId = aitem.SourceNodeId
+            aitem2.ReferenceTypeId = aitem.ReferenceTypeId
+            aitem2.IsForward = not forward
+            params.append(aitem2)
+
+        results = self.server.add_references(params)
+        for r in results:
+            r.check()
+
+    def set_modelling_rule(self, mandatory):
+        """
+        Add a modelling rule reference to Node.
+        When creating a new object type, its variable and child nodes will not
+        be instanciated if they do not have modelling rule
+        if mandatory is None, the modelling rule is removed
+        """
+        # remove all existing modelling rule
+        rules = self.get_references(ua.ObjectIds.HasModellingRule)
+        self.server.delete_references(list(map(self._fill_delete_reference_item, rules)))
+        # add new modelling rule as requested
+        if mandatory is not None:
+            rule = ua.ObjectIds.ModellingRule_Mandatory if mandatory else ua.ObjectIds.ModellingRule_Optional
+            self.add_reference(rule, ua.ObjectIds.HasModellingRule, True, False)
 
     def add_folder(self, nodeid, bname):
         return  opcua.common.manage_nodes.create_folder(self, nodeid, bname)
@@ -594,8 +673,8 @@ class Node(object):
     def add_method(self, *args):
         return opcua.common.manage_nodes.create_method(self, *args)
 
-    def add_reference_type(self, parent, nodeid, bname):
-        return opcua.common.manage_nodes.create_reference_type(parent, nodeid, bname)
+    def add_reference_type(self, nodeid, bname, symmetric=True, inversename=None):
+        return opcua.common.manage_nodes.create_reference_type(self, nodeid, bname, symmetric, inversename)
 
     def call_method(self, methodid, *args):
         return opcua.common.methods.call_method(self, methodid, *args)
