@@ -6,10 +6,13 @@ import logging
 from collections import OrderedDict
 import xml.etree.ElementTree as Et
 from copy import copy
+import base64
 
 from opcua import ua
 from opcua.ua import object_ids as o_ids
 from opcua.common.ua_utils import get_base_data_type
+from opcua.ua.uatypes import extension_object_ids
+from opcua.ua.uaerrors import UaError
 
 
 class XmlExporter(object):
@@ -92,7 +95,12 @@ class XmlExporter(object):
         idxs = []
         for node in nodes:
             node_idxs = [node.nodeid.NamespaceIndex]
-            node_idxs.append(node.get_browse_name().NamespaceIndex)
+            try:
+                node_idxs.append(node.get_browse_name().NamespaceIndex)
+            except UaError:
+                self.logger.exception("Error retrieving browse name of node %s", node)
+                raise
+
             node_idxs.extend(ref.NodeId.NamespaceIndex for ref in node.get_references())
             node_idxs = list(set(node_idxs))  # remove duplicates
             for i in node_idxs:
@@ -121,7 +129,7 @@ class XmlExporter(object):
         # from IPython import embed
         # embed()
         if pretty:
-            self.indent(self.etree.getroot())
+            indent(self.etree.getroot())
             self.etree.write(xmlpath,
                              encoding='utf-8',
                              xml_declaration=True
@@ -191,7 +199,7 @@ class XmlExporter(object):
         browsename = node.get_browse_name()
         nodeid = node.nodeid
         parent = node.get_parent()
-        displayname = node.get_display_name().Text.decode('utf-8')
+        displayname = node.get_display_name().Text
         desc = node.get_description().Text
         node_el = Et.SubElement(self.etree.getroot(), nodetype)
         node_el.attrib["NodeId"] = self._node_to_string(nodeid)
@@ -202,7 +210,7 @@ class XmlExporter(object):
                 node_el.attrib["ParentNodeId"] = self._node_to_string(parent)
         self._add_sub_el(node_el, 'DisplayName', displayname)
         if desc not in (None, ""):
-            self._add_sub_el(node_el, 'Description', desc.decode('utf-8'))
+            self._add_sub_el(node_el, 'Description', desc)
         # FIXME: add WriteMask and UserWriteMask
         return node_el
 
@@ -297,7 +305,7 @@ class XmlExporter(object):
         self._add_ref_els(obj_el, obj)
         var = obj.get_attribute(ua.AttributeIds.InverseName)
         if var is not None and var.Value.Value is not None and var.Value.Value.Text is not None:
-            self._add_sub_el(obj_el, 'InverseName', var.Value.Value.Text.decode('utf-8'))
+            self._add_sub_el(obj_el, 'InverseName', var.Value.Value.Text)
 
     def add_etree_datatype(self, obj):
         """
@@ -355,8 +363,6 @@ class XmlExporter(object):
 
 
     def _val_to_etree(self, el, dtype, val):
-        if val is None:
-            val = ""
         if dtype == ua.NodeId(ua.ObjectIds.NodeId):
             id_el = Et.SubElement(el, "uax:Identifier")
             id_el.text = val.to_string()
@@ -365,13 +371,20 @@ class XmlExporter(object):
             id_el.text = str(val)
         elif dtype == ua.NodeId(ua.ObjectIds.Boolean):
             el.text = 'true' if val else 'false'
+        elif dtype == ua.NodeId(ua.ObjectIds.ByteString):
+            if val is None:
+                val = b""
+            data = base64.b64encode(val)
+            el.text = data.decode("utf-8")
         elif not hasattr(val, "ua_types"):
             if isinstance(val, bytes):
+                # FIXME: should we also encode this (localized text I guess) using base64??
                 el.text = val.decode("utf-8")
             else:
-                el.text = str(val)
+                if val is not None:
+                    el.text = str(val)
         else:
-            for name, vtype in val.ua_types.items():
+            for name, vtype in val.ua_types:
                 self.member_to_etree(el, name, ua.NodeId(getattr(ua.ObjectIds, vtype)), getattr(val, name))
 
 
@@ -381,7 +394,6 @@ class XmlExporter(object):
             val_el = Et.SubElement(el, 'Value')
             self._value_to_etree(val_el, dtype_name, dtype, var.Value)
 
-
     def _value_to_etree(self, el, type_name, dtype, val):
         if val is None:
             return
@@ -411,6 +423,8 @@ class XmlExporter(object):
                 self._extobj_to_etree(el, type_name, dtype, val)
 
 
+
+
     def _extobj_to_etree(self, val_el, name, dtype, val):
         obj_el = Et.SubElement(val_el, "uax:ExtensionObject")
         type_el = Et.SubElement(obj_el, "uax:TypeId")
@@ -418,8 +432,14 @@ class XmlExporter(object):
         id_el.text = dtype.to_string()
         body_el = Et.SubElement(obj_el, "uax:Body")
         struct_el = Et.SubElement(body_el, "uax:" + name)
-        for name in self._get_member_order(dtype, val):
-            self.member_to_etree(struct_el, name, ua.NodeId(getattr(ua.ObjectIds, val.ua_types[name])), getattr(val, name))
+        for name, vtype  in val.ua_types:
+            # FIXME; what happend if we have a custom type which is not part of ObjectIds???
+            if vtype.startswith("ListOf"):
+                vtype = vtype[6:]
+            self.member_to_etree(struct_el, name, ua.NodeId(getattr(ua.ObjectIds, vtype)), getattr(val, name))
+            #self.member_to_etree(struct_el, name, extension_object_ids[vtype], getattr(val, name))
+        #for name in self._get_member_order(dtype, val):
+            #self.member_to_etree(struct_el, name, ua.NodeId(getattr(ua.ObjectIds, val.ua_types[name])), getattr(val, name))
 
     def _get_member_order(self, dtype, val):
         '''
@@ -433,22 +453,22 @@ class XmlExporter(object):
 
         return member_keys
 
-    def indent(self, elem, level=0):
-        """
-        copy and paste from http://effbot.org/zone/element-lib.htm#prettyprint
-        it basically walks your tree and adds spaces and newlines so the tree is
-        printed in a nice way
-        """
-        i = "\n" + level * "  "
-        if len(elem):
-            if not elem.text or not elem.text.strip():
-                elem.text = i + "  "
-            if not elem.tail or not elem.tail.strip():
-                elem.tail = i
-            for elem in elem:
-                self.indent(elem, level + 1)
-            if not elem.tail or not elem.tail.strip():
-                elem.tail = i
-        else:
-            if level and (not elem.tail or not elem.tail.strip()):
-                elem.tail = i
+def indent(elem, level=0):
+    """
+    copy and paste from http://effbot.org/zone/element-lib.htm#prettyprint
+    it basically walks your tree and adds spaces and newlines so the tree is
+    printed in a nice way
+    """
+    i = "\n" + level * "  "
+    if len(elem):
+        if not elem.text or not elem.text.strip():
+            elem.text = i + "  "
+        if not elem.tail or not elem.tail.strip():
+            elem.tail = i
+        for elem in elem:
+            indent(elem, level + 1)
+        if not elem.tail or not elem.tail.strip():
+            elem.tail = i
+    else:
+        if level and (not elem.tail or not elem.tail.strip()):
+            elem.tail = i
