@@ -9,6 +9,7 @@ from copy import copy
 import opcua
 from opcua import ua
 from opcua.common import xmlparser
+from opcua.ua.uaerrors import UaError
 
 import sys
 
@@ -23,6 +24,7 @@ class XmlImporter(object):
         self.server = server
         self.namespaces = {}
         self.aliases = {}
+        self.refs = None
 
     def _map_namespaces(self, namespaces_uris):
         """
@@ -37,23 +39,24 @@ class XmlImporter(object):
 
     def _map_aliases(self, aliases):
         """
-        maps the import aliases to the correct namespaces        
+        maps the import aliases to the correct namespaces
         """
         aliases_mapped = {}
         for alias, node_id in aliases.items():
-            aliases_mapped[alias] = self._migrate_ns(self.to_nodeid(node_id))
+            aliases_mapped[alias] = self.to_nodeid(node_id)
         return aliases_mapped
 
-    def import_xml(self, xmlpath):
+    def import_xml(self, xmlpath=None, xmlstring=None):
         """
         import xml and return added nodes
         """
         self.logger.info("Importing XML file %s", xmlpath)
-        self.parser = xmlparser.XMLParser(xmlpath)
+        self.parser = xmlparser.XMLParser(xmlpath, xmlstring)
 
         self.namespaces = self._map_namespaces(self.parser.get_used_namespaces())
         self.aliases = self._map_aliases(self.parser.get_aliases())
-        
+        self.refs = []
+
         dnodes = self.parser.get_node_datas()
         dnodes = self.make_objects(dnodes)
         nodes_parsed = self._sort_nodes_by_parentid(dnodes)
@@ -66,6 +69,12 @@ class XmlImporter(object):
                 self.logger.warning("failure adding node %s", nodedata)
                 raise
             nodes.append(node)
+
+        self.refs, remaining_refs = [], self.refs
+        self._add_references(remaining_refs)
+        if len(self.refs) != 0:
+            self.logger.warning("The following references could not be imported and are probaly broken: %s", self.refs) 
+
         return nodes
 
     def _add_node_data(self, nodedata):
@@ -95,10 +104,14 @@ class XmlImporter(object):
 
     def _add_references(self, refs):
         if isinstance(self.server, opcua.server.server.Server):
-            return self.server.iserver.isession.add_references(refs)
+            res = self.server.iserver.isession.add_references(refs)
         else:
-            return self.server.uaclient.add_references(refs)
-    
+            res = self.server.uaclient.add_references(refs)
+
+        for sc, ref in zip(res, refs):
+            if not sc.is_good():
+                self.refs.append(ref)
+
     def make_objects(self, node_datas):
         new_nodes = []
         for ndata in node_datas:
@@ -132,15 +145,14 @@ class XmlImporter(object):
         node.BrowseName = self._migrate_ns(obj.browsename)
         self.logger.info("Importing xml node (%s, %s) as (%s %s)", obj.browsename, obj.nodeid, node.BrowseName, node.RequestedNewNodeId)
         node.NodeClass = getattr(ua.NodeClass, obj.nodetype[2:])
-        if obj.parent:
+        if obj.parent and obj.parentlink:
             node.ParentNodeId = self._migrate_ns(obj.parent)
-        if obj.parentlink:
             node.ReferenceTypeId = self._migrate_ns(obj.parentlink)
         if obj.typedef:
             node.TypeDefinition = self._migrate_ns(obj.typedef)
         return node
 
-    def to_nodeid(self, nodeid):
+    def _to_nodeid(self, nodeid):
         if isinstance(nodeid, ua.NodeId):
             return nodeid
         elif not nodeid:
@@ -155,6 +167,9 @@ class XmlImporter(object):
             else:
                 return ua.NodeId(getattr(ua.ObjectIds, nodeid))
 
+    def to_nodeid(self, nodeid):
+        return self._migrate_ns(self._to_nodeid(nodeid))
+
     def add_object(self, obj):
         node = self._get_node(obj)
         attrs = ua.ObjectAttributes()
@@ -206,22 +221,41 @@ class XmlImporter(object):
         res[0].StatusCode.check()
         return res[0].AddedNodeId
 
+    def _get_ext_class(self, name):
+        if hasattr(ua, name):
+            return  getattr(ua, name)
+        elif name in self.aliases.keys():
+            nodeid = self.aliases[name]
+            class_type = ua.uatypes.get_extensionobject_class_type(nodeid)
+            if class_type:
+                return class_type
+            else:
+                raise Exception("Error no extension class registered ", name, nodeid)
+        else:
+            raise Exception("Error no alias found for extension class", name)
+
     def _make_ext_obj(self, obj):
-        ext = getattr(ua, obj.objname)()
+        ext = self._get_ext_class(obj.objname)()
         for name, val in obj.body:
-            if isinstance(val, str):
-                raise Exception("Error val should a dict", name, val)
+            if not isinstance(val, list):
+                raise Exception("Error val should be a list, this is a python-opcua bug", name, type(val), val)
             else:
                 for attname, v in val:
                     self._set_attr(ext, attname, v)
         return ext
 
+    def _get_val_type(self, obj, attname):
+        for name, uatype in obj.ua_types:
+            if name == attname:
+                return uatype
+        raise UaError("Attribute '{}' defined in xml is not found in object '{}'".format(attname, ext))
+
     def _set_attr(self, obj, attname, val):
         # tow possible values:
         # either we get value directly
         # or a dict if it s an object or a list
         if isinstance(val, (str, unicode)):
-            pval = xmlparser.ua_type_to_python(val, obj.ua_types[attname])
+            pval = xmlparser.ua_type_to_python(val, self._get_val_type(obj, attname))
             setattr(obj, attname, pval)
         else:
             # so we have either an object or a list...
@@ -229,10 +263,13 @@ class XmlImporter(object):
             if isinstance(obj2, ua.NodeId):  # NodeId representation does not follow common rules!!
                 for attname2, v2 in val:
                     if attname2 == "Identifier":
-                        obj2 = ua.NodeId.from_string(v2)
-                        setattr(obj, attname, obj2)
+                        if hasattr(ua.ObjectIds, v2):
+                            obj2 = ua.NodeId(getattr(ua.ObjectIds, v2))
+                        else:
+                            obj2 = ua.NodeId.from_string(v2)
+                        setattr(obj, attname, self._migrate_ns(obj2))
                         break
-            elif not isinstance(obj2, ua.NodeId) and not hasattr(obj2, "ua_types"):
+            elif not hasattr(obj2, "ua_types"):
                 # we probably have a list
                 my_list = []
                 for vtype, v2 in val:
@@ -273,7 +310,7 @@ class XmlImporter(object):
             ltext = ua.LocalizedText()
             for name, val in obj.value:
                 if name == "Text":
-                    ltext.Text = val.encode("utf-8")
+                    ltext.Text = val
                 else:
                     self.logger.warning("While parsing localizedText value, unkown element: %s with val: %s", name, val)
             return ua.Variant(ltext, ua.VariantType.LocalizedText)
@@ -361,11 +398,10 @@ class XmlImporter(object):
         refs = []
         for data in obj.refs:
             ref = ua.AddReferencesItem()
-            ref.IsForward = True
+            ref.IsForward = data.forward
             ref.ReferenceTypeId = self.to_nodeid(data.reftype)
             ref.SourceNodeId = self._migrate_ns(obj.nodeid)
-            ref.TargetNodeClass = ua.NodeClass.DataType
-            ref.TargetNodeId = self._migrate_ns(self.to_nodeid(data.target))
+            ref.TargetNodeId = self.to_nodeid(data.target)
             refs.append(ref)
         self._add_references(refs)
 
