@@ -5,6 +5,7 @@ from threading import Lock
 import sqlite3
 
 from opcua import ua
+from opcua.ua.ua_binary import variant_from_binary, variant_to_binary
 from opcua.common.utils import Buffer
 from opcua.common import events
 from opcua.server.history import HistoryStorageInterface
@@ -66,7 +67,7 @@ class HistorySQLite(HistoryStorageInterface):
                                    datavalue.StatusCode.value,
                                    str(datavalue.Value.Value),
                                    datavalue.Value.VariantType.name,
-                                   sqlite3.Binary(datavalue.Value.to_binary())
+                                   sqlite3.Binary(variant_to_binary(datavalue.Value))
                                )
                               )
             except sqlite3.Error as e:
@@ -90,12 +91,12 @@ class HistorySQLite(HistoryStorageInterface):
             if period:
                 # after the insert, if a period was specified delete all records older than period
                 date_limit = datetime.utcnow() - period
-                execute_sql_delete('ServerTimestamp < ?', (date_limit,))
+                execute_sql_delete('SourceTimestamp < ?', (date_limit,))
 
             if count:
                 # ensure that no more than count records are stored for the specified node
-                execute_sql_delete('ServerTimestamp = (SELECT CASE WHEN COUNT(*) > ? '
-                                   'THEN MIN(ServerTimestamp) ELSE NULL END FROM "{tn}")', (count,))
+                execute_sql_delete('SourceTimestamp = (SELECT CASE WHEN COUNT(*) > ? '
+                                   'THEN MIN(SourceTimestamp) ELSE NULL END FROM "{tn}")', (count,))
 
     def read_node_history(self, node_id, start, end, nb_values):
         with self._lock:
@@ -109,12 +110,12 @@ class HistorySQLite(HistoryStorageInterface):
 
             # select values from the database; recreate UA Variant from binary
             try:
-                for row in _c_read.execute('SELECT * FROM "{tn}" WHERE "ServerTimestamp" BETWEEN ? AND ? '
+                for row in _c_read.execute('SELECT * FROM "{tn}" WHERE "SourceTimestamp" BETWEEN ? AND ? '
                                            'ORDER BY "_Id" {dir} LIMIT ?'.format(tn=table, dir=order),
                                            (start_time, end_time, limit,)):
 
                     # rebuild the data value object
-                    dv = ua.DataValue(ua.Variant.from_binary(Buffer(row[6])))
+                    dv = ua.DataValue(variant_from_binary(Buffer(row[6])))
                     dv.ServerTimestamp = row[1]
                     dv.SourceTimestamp = row[2]
                     dv.StatusCode = ua.StatusCode(row[3])
@@ -126,7 +127,7 @@ class HistorySQLite(HistoryStorageInterface):
 
             if nb_values:
                 if len(results) > nb_values:
-                    cont = results[nb_values].ServerTimestamp
+                    cont = results[nb_values].SourceTimestamp
 
                 results = results[:nb_values]
 
@@ -215,7 +216,7 @@ class HistorySQLite(HistoryStorageInterface):
                     cont_timestamps.append(row[0])
                     for i, field in enumerate(row[1:]):
                         if field is not None:
-                            fdict[clauses[i]] = ua.Variant.from_binary(Buffer(field))
+                            fdict[clauses[i]] = variant_from_binary(Buffer(field))
                         else:
                             fdict[clauses[i]] = ua.Variant(None)
 
@@ -251,7 +252,7 @@ class HistorySQLite(HistoryStorageInterface):
 
         ev_fields = []
         for field in set(ev_aggregate_fields):
-            ev_fields.append(field.get_display_name().Text.decode(encoding='utf-8'))
+            ev_fields.append(field.get_display_name().Text)
         return ev_fields
 
     @staticmethod
@@ -302,7 +303,7 @@ class HistorySQLite(HistoryStorageInterface):
         for name in names:
             variant = ev_variant_dict[name]
             placeholders.append('?')
-            ev_variant_binaries.append(sqlite3.Binary(variant.to_binary()))
+            ev_variant_binaries.append(sqlite3.Binary(variant_to_binary(variant)))
 
         return self._list_to_sql_str(names), self._list_to_sql_str(placeholders, False), tuple(ev_variant_binaries)
 
