@@ -195,7 +195,7 @@ class MonitoredItemService(object):
                 mid = self._monitored_datachange[handle]
                 mdata = self._monitored_items[mid]
                 mdata.mvalue.set_current_value(value.Value.Value)
-                if mdata.filter is not None:
+                if mdata.filter:
                     deadband_flag_pass = self.deadband_callback(mdata.mvalue, mdata.filter)
                 else:
                     deadband_flag_pass = True
@@ -205,7 +205,6 @@ class MonitoredItemService(object):
                     self.isub.enqueue_datachange_event(mid, event, mdata.queue_size)
 
     def deadband_callback(self, values, flt):
-        ua.DeadbandType.None_
         if flt.DeadbandType == ua.DeadbandType.None_ or values.get_old_value() is None:
             return True
         elif flt.DeadbandType == ua.DeadbandType.Absolute and \
@@ -219,13 +218,13 @@ class MonitoredItemService(object):
 
     def trigger_event(self, event):
         with self._lock:
-            if event.SourceNode not in self._monitored_events:
+            if event.emitting_node not in self._monitored_events:
                 self.logger.debug("%s has no subscription for events %s from node: %s",
-                                  self, event, event.SourceNode)
+                                  self, event, event.emitting_node)
                 return False
             self.logger.debug("%s has subscription for events %s from node: %s",
-                              self, event, event.SourceNode)
-            mids = self._monitored_events[event.SourceNode]
+                              self, event, event.emitting_node)
+            mids = self._monitored_events[event.emitting_node]
             for mid in mids:
                 self._trigger_event(event, mid)
 
@@ -273,13 +272,18 @@ class InternalSubscription(object):
 
     def start(self):
         self.logger.debug("starting subscription %s", self.data.SubscriptionId)
-        self._subscription_loop()
+        if self.data.RevisedPublishingInterval > 0.0:
+            self._subscription_loop()
 
     def stop(self):
         self.logger.debug("stopping subscription %s", self.data.SubscriptionId)
         self._stopev = True
         self.monitored_item_srv.delete_all_monitored_items()
 
+    def _trigger_publish(self):
+        if not self._stopev and self.data.RevisedPublishingInterval <= 0.0:
+            self.subservice.loop.call_soon(self.publish_results)
+
     def _subscription_loop(self):
         if not self._stopev:
             self.subservice.loop.call_later(self.data.RevisedPublishingInterval / 1000.0, self._sub_loop)
@@ -361,15 +365,15 @@ class InternalSubscription(object):
         with self._lock:
             self._publish_cycles_count = 0
             for nb in acks:
-                if nb in self._not_acknowledged_results:
-                    self._not_acknowledged_results.pop(nb)
+                self._not_acknowledged_results.pop(nb, None)
 
     def republish(self, nb):
         self.logger.info("re-publish request for ack %s in subscription %s", nb, self)
         with self._lock:
-            if nb in self._not_acknowledged_results:
+            notificationMessage = self._not_acknowledged_results.pop(nb, None)
+            if notificationMessage:
                 self.logger.info("re-publishing ack %s in subscription %s", nb, self)
-                return self._not_acknowledged_results[nb].NotificationMessage
+                return notificationMessage
             else:
                 self.logger.info("Error request to re-published non existing ack %s in subscription %s", nb, self)
                 return ua.NotificationMessage()
@@ -382,15 +386,18 @@ class InternalSubscription(object):
 
     def enqueue_statuschange(self, code):
         self._triggered_statuschanges.append(code)
+        self._trigger_publish()
 
     def _enqueue_event(self, mid, eventdata, size, queue):
-        if mid not in queue:
-            queue[mid] = [eventdata]
-            return
-        if size != 0:
-            if len(queue[mid]) >= size:
-                queue[mid].pop(0)
-        queue[mid].append(eventdata)
+        with self._lock:
+            if mid not in queue:
+                queue[mid] = [eventdata]
+                self._trigger_publish()
+                return
+            if size != 0:
+                if len(queue[mid]) >= size:
+                    queue[mid].pop(0)
+            queue[mid].append(eventdata)
 
 
 class WhereClauseEvaluator(object):
