@@ -3,10 +3,14 @@ import unittest
 from opcua import Client
 from opcua import Server
 from opcua import ua
+from opcua.client.ua_client import UASocketClient
+from opcua.common.utils import SocketWrapper
 
 from tests_subscriptions import SubscriptionTests
 from tests_common import CommonTests, add_server_methods
-from tests_xml import XmlTests 
+from tests_xml import XmlTests
+
+from tests_enum_struct import add_server_custom_enum_struct
 
 port_num1 = 48510
 
@@ -22,18 +26,19 @@ class TestClient(unittest.TestCase, CommonTests, SubscriptionTests, XmlTests):
     def setUpClass(cls):
         # start our own server
         cls.srv = Server()
-        cls.srv.set_endpoint('opc.tcp://localhost:{0:d}'.format(port_num1))
+        cls.srv.set_endpoint('opc.tcp://127.0.0.1:{0:d}'.format(port_num1))
         add_server_methods(cls.srv)
+        add_server_custom_enum_struct(cls.srv)
         cls.srv.start()
 
         # start admin client
         # long timeout since travis (automated testing) can be really slow
-        cls.clt = Client('opc.tcp://admin@localhost:{0:d}'.format(port_num1), timeout=10)
+        cls.clt = Client('opc.tcp://admin@127.0.0.1:{0:d}'.format(port_num1), timeout=10)
         cls.clt.connect()
         cls.opc = cls.clt
 
         # start anonymous client
-        cls.ro_clt = Client('opc.tcp://localhost:{0:d}'.format(port_num1))
+        cls.ro_clt = Client('opc.tcp://127.0.0.1:{0:d}'.format(port_num1))
         cls.ro_clt.connect()
 
     @classmethod
@@ -99,3 +104,32 @@ class TestClient(unittest.TestCase, CommonTests, SubscriptionTests, XmlTests):
             self.assertEqual(state[0], 1)
         # test if client disconnected
         self.assertEqual(state[0], 2)
+
+    def test_enumstrings_getvalue(self):
+        ''' The real exception is server side, but is detected by using a client.
+            Alldue the server trace is also visible on the console.
+            The client only 'sees' an TimeoutError
+        '''
+        nenumstrings = self.clt.get_node(ua.ObjectIds.AxisScaleEnumeration_EnumStrings)
+        with self.assertNotRaises(Exception):
+            value = ua.Variant(nenumstrings.get_value())
+
+    def test_uasocketclient_connect_disconnect(self):
+        """Initialize, connect, and disconnect a UaSocketClient
+        """
+        uaclt = UASocketClient()
+        uaclt.connect_socket('127.0.0.1', port_num1)
+        self.assertTrue(uaclt._thread.is_alive())
+        self.assertIsInstance(uaclt._socket, SocketWrapper)
+
+        # disconnect_socket() should shut down the receiving thread
+        uaclt.disconnect_socket()
+        self.assertFalse(uaclt._thread.is_alive())
+
+    def test_custom_enum_struct(self):
+        self.ro_clt.load_type_definitions()
+        ns = self.ro_clt.get_namespace_index('http://yourorganisation.org/struct_enum_example/')
+        myvar = self.ro_clt.get_node(ua.NodeId(6009, ns))
+        val = myvar.get_value()
+        self.assertEqual(val.IntVal1, 242)
+        self.assertEqual(val.EnumVal, ua.ExampleEnum.EnumVal2)
