@@ -4,6 +4,7 @@ import time
 from datetime import datetime
 from datetime import timedelta
 import math
+from contextlib import contextmanager
 
 from opcua import ua
 from opcua import Node
@@ -11,7 +12,7 @@ from opcua import uamethod
 from opcua import instantiate
 from opcua import copy_node
 from opcua.common import ua_utils
-
+from opcua.common.methods import call_method_full
 
 def add_server_methods(srv):
     @uamethod
@@ -23,6 +24,13 @@ def add_server_methods(srv):
 
     @uamethod
     def func2(parent, methodname, value):
+        if methodname == "panic":
+            return ua.StatusCode(ua.StatusCodes.BadOutOfMemory)
+        if methodname != "sin":
+            res = ua.CallMethodResult()
+            res.StatusCode = ua.StatusCode(ua.StatusCodes.BadInvalidArgument)
+            res.InputArgumentResults = [ua.StatusCode(ua.StatusCodes.BadNotSupported), ua.StatusCode()]
+            return res
         return math.sin(value)
 
     o = srv.get_objects_node()
@@ -35,6 +43,25 @@ def add_server_methods(srv):
     o = srv.get_objects_node()
     v = o.add_method(ua.NodeId("ServerMethodArray2", 2), ua.QualifiedName('ServerMethodArray2', 2), func3, [ua.VariantType.Int64], [ua.VariantType.Int64])
 
+    @uamethod
+    def func4(parent):
+        return None
+
+    base_otype= srv.get_node(ua.ObjectIds.BaseObjectType)
+    custom_otype = base_otype.add_object_type(2, 'ObjectWithMethodsType')
+    custom_otype.add_method(2, 'ServerMethodDefault', func4)
+    custom_otype.add_method(2, 'ServerMethodMandatory', func4).set_modelling_rule(True)
+    custom_otype.add_method(2, 'ServerMethodOptional', func4).set_modelling_rule(False)
+    custom_otype.add_method(2, 'ServerMethodNone', func4).set_modelling_rule(None)
+    o.add_object(2, 'ObjectWithMethods', custom_otype)
+
+    @uamethod
+    def func5(parent):
+        return 1,2,3
+    o = srv.get_objects_node()
+    v = o.add_method(ua.NodeId("ServerMethodTuple", 2), ua.QualifiedName('ServerMethodTuple', 2), func5, [], [ua.VariantType.Int64, ua.VariantType.Int64, ua.VariantType.Int64])
+
+
 
 class CommonTests(object):
 
@@ -48,6 +75,13 @@ class CommonTests(object):
     assertEqual = lambda x, y: True
     assertIn = lambda x, y: True
 
+    @contextmanager
+    def assertNotRaises(self, exc_type):
+        try:
+            yield None
+        except exc_type:
+            raise self.failureException('{} raised'.format(exc_type.__name__))
+
     def test_find_servers(self):
         servers = self.opc.find_servers()
         # FIXME : finish
@@ -105,15 +139,72 @@ class CommonTests(object):
             var = fold.add_property(2, "ProToDeleteR", 9.1)
             prop = fold.add_property(2, "ProToDeleteR", 9.1)
             o = fold.add_object(3, "ObjToDeleteR")
+            o_var = o.add_variable(3, "VarToDeleteRR", 9.2)
+            o_prop = o.add_property(3, "PropToDeleteRR", 9.2)
             mynodes.append(nfold)
             mynodes.append(var)
             mynodes.append(prop)
             mynodes.append(o)
+            mynodes.append(o_var)
+            mynodes.append(o_prop)
         self.opc.delete_nodes([fold], recursive=True)
         for node in mynodes:
             with self.assertRaises(ua.UaStatusCodeError):
                 node.get_browse_name()
 
+    def test_delete_references(self):
+        newtype = self.opc.get_node(ua.ObjectIds.HierarchicalReferences).add_reference_type(0, "HasSuperSecretVariable")
+
+        obj = self.opc.get_objects_node()
+        fold = obj.add_folder(2, "FolderToRef")
+        var = fold.add_variable(2, "VarToRef", 42)
+
+        fold.add_reference(var, newtype)
+
+        self.assertEqual(var.get_referenced_nodes(newtype), [fold])
+        self.assertEqual(fold.get_referenced_nodes(newtype), [var])
+
+        fold.delete_reference(var, newtype)
+
+        self.assertEqual(var.get_referenced_nodes(newtype), [])
+        self.assertEqual(fold.get_referenced_nodes(newtype), [])
+
+        fold.add_reference(var, newtype, bidirectional=False)
+
+        self.assertEqual(var.get_referenced_nodes(newtype), [])
+        self.assertEqual(fold.get_referenced_nodes(newtype), [var])
+
+        fold.delete_reference(var, newtype)
+
+        self.assertEqual(var.get_referenced_nodes(newtype), [])
+        self.assertEqual(fold.get_referenced_nodes(newtype), [])
+
+
+        var.add_reference(fold, newtype, forward=False, bidirectional=False)
+
+        self.assertEqual(var.get_referenced_nodes(newtype), [fold])
+        self.assertEqual(fold.get_referenced_nodes(newtype), [])
+
+        with self.assertRaises(ua.UaStatusCodeError):
+            fold.delete_reference(var, newtype)
+
+        self.assertEqual(var.get_referenced_nodes(newtype), [fold])
+        self.assertEqual(fold.get_referenced_nodes(newtype), [])
+
+        with self.assertRaises(ua.UaStatusCodeError):
+            var.delete_reference(fold, newtype)
+
+        self.assertEqual(var.get_referenced_nodes(newtype), [fold])
+        self.assertEqual(fold.get_referenced_nodes(newtype), [])
+
+        var.delete_reference(fold, newtype, forward=False)
+
+        self.assertEqual(var.get_referenced_nodes(newtype), [])
+        self.assertEqual(fold.get_referenced_nodes(newtype), [])
+
+        # clean-up
+        self.opc.delete_nodes([fold, newtype], recursive=True)
+
     def test_server_node(self):
         node = self.opc.get_server_node()
         self.assertEqual(ua.QualifiedName('Server', 0), node.get_browse_name())
@@ -453,12 +544,37 @@ class CommonTests(object):
         m = o.get_child("2:ServerMethodArray")
         result = o.call_method(m, "sin", ua.Variant(math.pi))
         self.assertTrue(result < 0.01)
+        with self.assertRaises(ua.UaStatusCodeError) as cm:
+            result = o.call_method(m, "cos", ua.Variant(math.pi))
+        self.assertEqual(cm.exception.code, ua.StatusCodes.BadInvalidArgument)
+        with self.assertRaises(ua.UaStatusCodeError) as cm:
+            result = o.call_method(m, "panic", ua.Variant(math.pi))
+        self.assertEqual(cm.exception.code, ua.StatusCodes.BadOutOfMemory)
 
     def test_method_array2(self):
         o = self.opc.get_objects_node()
         m = o.get_child("2:ServerMethodArray2")
         result = o.call_method(m, [1.1, 3.4, 9])
         self.assertEqual(result, [2.2, 6.8, 18])
+        result = call_method_full(o, m, [1.1, 3.4, 9])
+        self.assertEqual(result.OutputArguments, [[2.2, 6.8, 18]])
+
+    def test_method_tuple(self):
+        o = self.opc.get_objects_node()
+        m = o.get_child("2:ServerMethodTuple")
+        result = o.call_method(m)
+        self.assertEqual(result, [1, 2, 3])
+        result = call_method_full(o, m)
+        self.assertEqual(result.OutputArguments, [1, 2, 3])
+
+    def test_method_none(self):
+        # this test calls the function linked to the type's method..
+        o = self.opc.get_node(ua.ObjectIds.BaseObjectType).get_child("2:ObjectWithMethodsType")
+        m = o.get_child("2:ServerMethodDefault")
+        result = o.call_method(m)
+        self.assertEqual(result, None)
+        result = call_method_full(o, m)
+        self.assertEqual(result.OutputArguments, [])
 
     def test_add_nodes(self):
         objects = self.opc.get_objects_node()
@@ -479,6 +595,26 @@ class CommonTests(object):
         self.assertTrue(v in childs)
         self.assertTrue(p in childs)
 
+    def test_modelling_rules(self):
+        obj = self.opc.nodes.base_object_type.add_object_type(2, 'MyFooObjectType')
+        v = obj.add_variable(2, "myvar", 1.1)
+        v.set_modelling_rule(True)
+        p = obj.add_property(2, "myvar", 1.1)
+        p.set_modelling_rule(False)
+
+        refs = obj.get_referenced_nodes(refs=ua.ObjectIds.HasModellingRule)
+        self.assertEqual(len(refs), 0)
+
+        refs = v.get_referenced_nodes(refs=ua.ObjectIds.HasModellingRule)
+        self.assertEqual(refs[0], self.opc.get_node(ua.ObjectIds.ModellingRule_Mandatory))
+
+        refs = p.get_referenced_nodes(refs=ua.ObjectIds.HasModellingRule)
+        self.assertEqual(refs[0], self.opc.get_node(ua.ObjectIds.ModellingRule_Optional))
+
+        p.set_modelling_rule(None)
+        refs = p.get_referenced_nodes(refs=ua.ObjectIds.HasModellingRule)
+        self.assertEqual(len(refs), 0)
+
     def test_incl_subtypes(self):
         base_type = self.opc.get_root_node().get_child(["0:Types", "0:ObjectTypes", "0:BaseObjectType"])
         descs = base_type.get_children_descriptions(includesubtypes=True)
@@ -515,6 +651,7 @@ class CommonTests(object):
         self.assertTrue(objects in nodes)
         self.assertEqual(o.get_parent(), objects)
         self.assertEqual(o.get_type_definition().Identifier, ua.ObjectIds.BaseObjectType)
+        self.assertEqual(o.get_references(ua.ObjectIds.HasModellingRule), [])
 
         o2 = o.add_object(3, 'MySecondObject')
         nodes = o.get_referenced_nodes(refs=ua.ObjectIds.HasComponent, direction=ua.BrowseDirection.Forward, includesubtypes=False)
@@ -523,6 +660,7 @@ class CommonTests(object):
         self.assertTrue(o in nodes)
         self.assertEqual(o2.get_parent(), o)
         self.assertEqual(o2.get_type_definition().Identifier, ua.ObjectIds.BaseObjectType)
+        self.assertEqual(o2.get_references(ua.ObjectIds.HasModellingRule), [])
 
         v = o.add_variable(3, 'MyVariable', 6)
         nodes = o.get_referenced_nodes(refs=ua.ObjectIds.HasComponent, direction=ua.BrowseDirection.Forward, includesubtypes=False)
@@ -531,6 +669,7 @@ class CommonTests(object):
         self.assertTrue(o in nodes)
         self.assertEqual(v.get_parent(), o)
         self.assertEqual(v.get_type_definition().Identifier, ua.ObjectIds.BaseDataVariableType)
+        self.assertEqual(v.get_references(ua.ObjectIds.HasModellingRule), [])
 
         p = o.add_property(3, 'MyProperty', 2)
         nodes = o.get_referenced_nodes(refs=ua.ObjectIds.HasProperty, direction=ua.BrowseDirection.Forward, includesubtypes=False)
@@ -539,16 +678,17 @@ class CommonTests(object):
         self.assertTrue(o in nodes)
         self.assertEqual(p.get_parent(), o)
         self.assertEqual(p.get_type_definition().Identifier, ua.ObjectIds.PropertyType)
+        self.assertEqual(p.get_references(ua.ObjectIds.HasModellingRule), [])
+
+        m = objects.get_child("2:ServerMethod")
+        self.assertEqual(m.get_references(ua.ObjectIds.HasModellingRule), [])
 
     def test_path_string(self):
         o = self.opc.nodes.objects.add_folder(1, "titif").add_object(3, "opath")
-        path = o.get_path_as_string()
+        path = o.get_path(as_string=True)
         self.assertEqual(["0:Root", "0:Objects", "1:titif", "3:opath"], path)
-        path = o.get_path_as_string(2)
+        path = o.get_path(2, as_string=True)
         self.assertEqual(["1:titif", "3:opath"], path)
-        path = self.opc.get_node("i=13387").get_path_as_string()
-        # FIXME this is wrong in our server! BaseObjectType is missing an inverse reference to its parent! seems xml definition is wrong
-        self.assertEqual(['0:BaseObjectType', '0:FolderType', '0:FileDirectoryType', '0:CreateDirectory'], path) 
 
     def test_path(self):
         of = self.opc.nodes.objects.add_folder(1, "titif")
@@ -559,8 +699,7 @@ class CommonTests(object):
         self.assertEqual([of, op], path)
         target = self.opc.get_node("i=13387")
         path = target.get_path()
-        # FIXME this is wrong in our server! BaseObjectType is missing an inverse reference to its parent! seems xml definition is wrong
-        self.assertEqual([self.opc.nodes.base_object_type, self.opc.nodes.folder_type, self.opc.get_node(ua.ObjectIds.FileDirectoryType), target], path)  
+        self.assertEqual([self.opc.nodes.root, self.opc.nodes.types, self.opc.nodes.object_types, self.opc.nodes.base_object_type, self.opc.nodes.folder_type, self.opc.get_node(ua.ObjectIds.FileDirectoryType), target], path)
 
     def test_get_endpoints(self):
         endpoints = self.opc.get_endpoints()
@@ -577,7 +716,7 @@ class CommonTests(object):
         devd_t = dev_t.add_object_type(0, "MyDeviceDervived")
         v_t = devd_t.add_variable(0, "childparam", 1.0)
         p_t = devd_t.add_property(0, "sensorx_id", "0340")
-        
+
 
         nodes = copy_node(self.opc.nodes.objects, dev_t)
         mydevice = nodes[0]
@@ -594,15 +733,25 @@ class CommonTests(object):
         # Create device type
         dev_t = self.opc.nodes.base_object_type.add_object_type(0, "MyDevice")
         v_t = dev_t.add_variable(0, "sensor", 1.0)
+        v_t.set_modelling_rule(True)
         p_t = dev_t.add_property(0, "sensor_id", "0340")
+        p_t.set_modelling_rule(True)
         ctrl_t = dev_t.add_object(0, "controller")
+        ctrl_t.set_modelling_rule(True)
+        v_opt_t = dev_t.add_variable(0, "vendor", 1.0)
+        v_opt_t.set_modelling_rule(False)
+        v_none_t = dev_t.add_variable(0, "model", 1.0)
+        v_none_t.set_modelling_rule(None)
         prop_t = ctrl_t.add_property(0, "state", "Running")
+        prop_t.set_modelling_rule(True)
 
         # Create device sutype
         devd_t = dev_t.add_object_type(0, "MyDeviceDervived")
         v_t = devd_t.add_variable(0, "childparam", 1.0)
+        v_t.set_modelling_rule(True)
         p_t = devd_t.add_property(0, "sensorx_id", "0340")
-        
+        p_t.set_modelling_rule(True)
+
         # instanciate device
         nodes = instantiate(self.opc.nodes.objects, dev_t, bname="2:Device0001")
         mydevice = nodes[0]
@@ -611,13 +760,18 @@ class CommonTests(object):
         self.assertEqual(mydevice.get_type_definition(), dev_t.nodeid)
         obj = mydevice.get_child(["0:controller"])
         prop = mydevice.get_child(["0:controller", "0:state"])
+        with self.assertRaises(ua.UaError):
+            mydevice.get_child(["0:controller", "0:vendor"])
+        with self.assertRaises(ua.UaError):
+            mydevice.get_child(["0:controller", "0:model"])
+
         self.assertEqual(prop.get_type_definition().Identifier, ua.ObjectIds.PropertyType)
         self.assertEqual(prop.get_value(), "Running")
         self.assertNotEqual(prop.nodeid, prop_t.nodeid)
-        
+
         # instanciate device subtype
         nodes = instantiate(self.opc.nodes.objects, devd_t, bname="2:Device0002")
-        mydevicederived = nodes[0] 
+        mydevicederived = nodes[0]
         prop1 = mydevicederived.get_child(["0:sensorx_id"])
         var1 = mydevicederived.get_child(["0:childparam"])
         var_parent = mydevicederived.get_child(["0:sensor"])
@@ -627,9 +781,13 @@ class CommonTests(object):
         # Create device type
         dev_t = self.opc.nodes.base_object_type.add_object_type(0, "MyDevice2")
         v_t = dev_t.add_variable(0, "sensor", 1.0)
+        v_t.set_modelling_rule(True)
         p_t = dev_t.add_property(0, "sensor_id", "0340")
+        p_t.set_modelling_rule(True)
         ctrl_t = dev_t.add_object(0, "controller")
+        ctrl_t.set_modelling_rule(True)
         prop_t = ctrl_t.add_property(0, "state", "Running")
+        prop_t.set_modelling_rule(True)
 
         # instanciate device
         nodes = instantiate(self.opc.nodes.objects, dev_t, nodeid=ua.NodeId("InstDevice", 2, ua.NodeIdType.String), bname="2:InstDevice")
@@ -658,7 +816,7 @@ class CommonTests(object):
         # create enum type
         enums = self.opc.get_root_node().get_child(["0:Types", "0:DataTypes", "0:BaseDataType", "0:Enumeration"])
         myenum_type = enums.add_data_type(0, "MyEnum")
-        es = myenum_type.add_variable(0, "EnumStrings", [ua.LocalizedText("String0"), ua.LocalizedText("String1"), ua.LocalizedText("String2")], ua.VariantType.LocalizedText) 
+        es = myenum_type.add_variable(0, "EnumStrings", [ua.LocalizedText("String0"), ua.LocalizedText("String1"), ua.LocalizedText("String2")], ua.VariantType.LocalizedText)
         #es.set_value_rank(1)
         # instantiate
         o = self.opc.get_objects_node()
@@ -711,13 +869,8 @@ class CommonTests(object):
             ua.ObjectIds.Structure: ua.VariantType.ExtensionObject,
             ua.ObjectIds.EnumValueType: ua.VariantType.ExtensionObject,
             ua.ObjectIds.Enumeration: ua.VariantType.Int32,  # enumeration
-            ua.ObjectIds.AttributeWriteMask: ua.VariantType.Int32,  # enumeration
+            ua.ObjectIds.AttributeWriteMask: ua.VariantType.UInt32,
             ua.ObjectIds.AxisScaleEnumeration: ua.VariantType.Int32  # enumeration
             }
         for dt, vdt in test_data.items():
             self.assertEqual(ua_utils.data_type_to_variant_type(self.opc.get_node(ua.NodeId(dt))), vdt)
-
-    def test_variant_intenum(self):
-        ase = ua.AxisScaleEnumeration(ua.AxisScaleEnumeration.Linear)  # Just pick an existing IntEnum class
-        vAse = ua.Variant(ase)
-        self.assertEqual(vAse.VariantType, ua.VariantType.Int32)
